/*

====parse_code(code)====

Returns new object of subclass parsed_code

Takes xml object code as output from ctoxml. Initializes
properties types, vars, enums, structs, unions, defs and funcs.

*/

C.parse_code=function(code) {
	var ret=new C.parsed_code;

	var decl;
	C.splitdeclarators(code); // Split declarators

	for each(decl in code.d) { // Fill up typedefs before pruning the tree
		var elem;
		var type=undefined;
		if (decl.typedef!=undefined) {
			type=ret.types;
			var ident=<a>{decl.*[decl.*.length()-1]}</a>..id[0];
			if (type!=undefined) type[ident]=decl;
		}
	}

	C.recursion(code); // Prune away deep includes

	for each(decl in code.d) {
		var ident=undefined;
		var elem;
		var type=undefined;
		if (decl.typedef!=undefined) { // typedef
			// here, only the ones not pruned away are kept
			type=ret.typedefs;
			ident=<a>{decl.*[decl.*.length()-1]}</a>..id[0];
		} else if (decl.typedef==undefined && decl.enum!=undefined) { // enum
			type=ret.enums;
			ident=decl.enum[0].@id;
		} else if (decl.typedef==undefined && (
			decl..fd!=undefined ||  // function
			decl.id!=undefined || // variable
			decl..id!=undefined)) { // indirect variable
			type=ret.ptrs;
			ident=decl.childIndex();
		}

		if (type!=undefined && ident==undefined) ident=C.getident(decl);
		if (type!=undefined && ident!=undefined)
			type[ident]=decl;
	}

	for each(decl in code.d) { // Struct, union and enum defs may be inside other structs.
		var ident;
		var elem;
		var type=undefined;
		if (decl.struct!=undefined && decl.struct.d!=undefined) {
			type=ret.structs;
			ident=decl.struct[0].@id;
		} else if (decl.union!=undefined && decl.union.d!=undefined) {
			type=ret.structs;
			ident=decl.union[0].@id;
		} else if (decl.enum!=undefined) {
			type=ret.enums;
			ident=decl.enum[0].@id;
		}

		if (type!=undefined && ident==undefined) ident=C.getident(decl);
		if (type!=undefined && ident!=undefined)
			if (type[ident]==undefined) type[ident]=decl;
	}

	// Find function definitions

	for each(decl in code.fdef) {
		ident=decl.childIndex(); //<a>{decl.*[decl.*.length()-2]}</a>..id[0];
		ret.ptrs[ident]=decl.copy();
		delete ret.ptrs[ident].p;
	}

	// Find macros and definitions

	for each(def in code.define) {
		ret.defs[def.id]=def;
	}

	// Find dls

	var ndl=0;
	var match;
	for each(pragma in code.pragma) {
		if ((match=pragma.match(/JSEXT[ \t]+dl[ \t]+(("([^"]*)")|(default)|(main))[ \t]*$/))) {
			if (match[3]) ret.dls[pragma.childIndex()]=match[3];
			else if (match[4]) ret.dls[pragma.childIndex()]=undefined;
			else if (match[5]) ret.dls[pragma.childIndex()]=0;
		}
	}

	return ret;
}

/*

====parsed_code()====

Constructor for subclass parsed_code

*/

C.parsed_code=function() {
	this.types={};
	this.typedefs={};
	this.ptrs={};
	this.enums={};
	this.structs={};
	this.defs={};
	this.macros={};
	this.dls={};
}

/*

====typestring(type)====

''decl'' should be an xml declaration.


Example...

typedef struct _IO_FILE FILE;
int fprintf(FILE *__stream, char *__format, ...);

Returns the string below:

   Identifier
   |     
   |    Pointer dimension
   |    |
   |    | Type (_F_unction, _s_truct, _u_nion, _f_loat, _i_nteger, _V_alist, _v_oid
   |    | |
   |    | |Return value
   |    | |
   |    | |Call convention (c = __cdecl, s = __stdcall)
   |    | ||
   |    | ||Identifier (none)
   |    | |||
   |    | |||Array dimensionality
   |    | ||||
   |    | |||| Type (integer)
   |    | |||| |
   |    | |||| |Size (0=char, 1=short, 2=int, 3=long, 4=long long)
   |    | |||| ||
   |    | |||| ||Signedness (_s_igned, _u_nsigned, ?=unspecified)
   |    | |||| |||
   |    | ||[====] Return value
   |    | |||| |||
   |    | |||| |||Number of arguments
   |    | |||| ||||
   |    | |||| |||| Const (0 or 1)
   |    | |||| |||| |
   |    | |||| |||| |Argument 1 type
   |    | |||| |||| |
   |    | |||| |||| |Identifier
   |    | |||| |||| |
   |    | |||| |||| |         Array dimensionality
   |    | |||| |||| |         |
   |    | |||| |||| |         | Array size, dimension 0
   |    | |||| |||| |         | |
   |    | |||| |||| |         | | Type (struct)
   |    | |||| |||| |         | | |
   |    | |||| |||| |         | | |Struct identifier
   |    | |||| |||| |         | | ||
   |    | |||| |||| |         | | ||
   |    | |||| |||| |         | | ||
   |    | |||| |||| |[=====================] Argument 1 type
   |    | |||| |||| |         | | ||
   |    | |||| |||| |         | | ||        Const (0 or 1)
   |    | |||| |||| |         | | ||        |
   |    | |||| |||| |         | | ||        |Argument 2 type
   |    | |||| |||| |         | | ||        |
   |    | |||| |||| |         | | ||        |Identifier
   |    | |||| |||| |         | | ||        ||
   |    | |||| |||| |         | | ||        ||        Array dimensionality        
   |    | |||| |||| |         | | ||        ||        |
   |    | |||| |||| |         | | ||        ||        | Array size, dimension 0
   |    | |||| |||| |         | | ||        ||        | |
   |    | |||| |||| |         | | ||        ||        | | Type (integer)
   |    | |||| |||| |         | | ||        ||        | | |
   |    | |||| |||| |         | | ||        ||        | | |Size (char)
   |    | |||| |||| |         | | ||        ||        | | ||
   |    | |||| |||| |         | | ||        ||        | | ||Signedness (unspecified)
   |    | |||| |||| |         | | ||        ||        | | |||
   |    | |||| |||| |         | | ||        |[==============] Argument 2 type
   |    | |||| |||| |         | | ||        ||        | | |||
   |    | |||| |||| |         | | ||        ||        | | |||Ellipsis (0 or 1)
   |    | |||| |||| |         | | ||        ||        | | ||||
   |    | |||| |||| |         | | ||        ||        | | ||||
fprintf 0 Fc 0 i2?2 0__stream 1 0 s_IO_FILE 1__format 1 0 i0?1


Floats and doubles are represented as

f0 (float), f1 (double) or f2 (long double)

*/

C.parsed_code.prototype.typestring=function(decl) {
	var dim=[];
	return recurse.call(this, this.type(decl,dim));

	function recurse(type) {
		var ret="";
		if (type.ident!=undefined) ret+=type.ident;
		ret+=" "+type.dim.length+" "+type.dim.join(" ")+(type.dim.length?" ":"");

		switch(type.type) {
		case "struct":
			if (type.struct_id) {
				return ret+"s"+type.struct_id+" ";
			} else {
				return ret+this.structstring(decl);
			}
		case "union":
			if (type.struct_id) {
				return ret+"u"+type.struct_id+" ";
			} else {
				return ret+this.structstring(decl);
			}
		case "enum":
			return ret+"i2?";
		case "float": return ret+"f"+type.size;
		case "int":
			ret+="i"+type.size;
			if (type.signed==undefined) ret+="?";
			else ret+=type.signed?"s":"u";
			return ret;
		case "__builtin_va_list": return ret+"V";
		case "void": return ret+"v";
		case "function":
			ret+="F";
			if (type.callconv=="stdcall") ret+="s";
			else ret+="c";

			if (type.rettype==undefined) ret+="v";
			else ret+=recurse.call(this,type.rettype);

			ret+=type.params.length+" ";
			var i;
			for (i in type.params) {
				ret+=type.params[i].isConst?1:0;
				ret+=recurse.call(this,type.params[i]);
			}
			ret+=type.elipsis?"1":"0";
			return ret;
		}
	}
}


/*

*/

C.parsed_code.prototype.enumvals=function(obj) {
	var ret={};
	var val=0;
	var prevsym=undefined;
	for each(var sym in obj.enum.*) {
		if (sym.name()=="c") val=Number(sym);
		if (sym.name()=="id") {
			if (prevsym!=undefined) {
				ret[prevsym]=val;
				val++;
			}
			prevsym=sym;
		}
	}
	if (prevsym!=undefined) ret[prevsym]=val;
	return ret;
}


/*

*/

C.parsed_code.prototype.structstring=function(decl) {
	var ret="";
	if (decl.struct!=undefined) {
		var decls=decl.struct.d;
		ret+="S";
	}
	if (decl.union!=undefined) {
		var decls=decl.union.d;
		ret+="U";
	}
	var nmembers=0;
	var membstring="";
	for each (var member in decls) {
		nmembers++;
		membstring+=this.typestring(member);
	}
	return ret+nmembers+" "+membstring;
}

C.parsed_code.prototype.type=function(decl, dim) {
	var type;
	var lasttype=undefined;
	var size=0;
	var signed=undefined;

	if (dim==undefined) dim=[];
	this.getdim(decl.*[decl.*.length()-1], dim);

	if (decl.dt!=undefined) {
	// Defined type
		type=this.type(this.types[decl.dt], dim);
		if (type.typename==undefined) {
			type.typename=String(decl.dt);
		}
	} else if (decl.struct!=undefined) {
	// Struct declaration
		type={type: "struct",
			struct_id: decl.struct.@id.toString(),
			dim: dim};
	} else if (decl.union!=undefined) {
	// Union declaration
		type={type: "union",
			struct_id: decl.union.@id.toString(),
			dim: dim};
	} else if (decl.enum!=undefined) {
	// Enum declaration
		type={type: "enum",
			dim: dim};
	} else {
	// Primitive declaration
		for each (type in decl.t) {
			if (type=="signed") signed=true;
			if (type=="unsigned") signed=false;
			if (type=="char") size-=2;
			if (type=="short") size--;
			if (type=="__int64") size=3;
			if (type=="long" || type=="double") size++;
			if (type=="int" ||
				type=="long" ||
				type=="char" ||
				type=="short" ||
				type=="__int64" ||
				type=="signed" ||
				type=="unsigned") lasttype="int";
			if (type=="float" ||
				type=="double") lasttype="float";
			if (type=="__builtin_va_list") lasttype="__builtin_va_list";
			if (type=="void") lasttype="void";
		}

		if (lasttype=="int") size+=2;

		type={type: lasttype,
			size: size,
			signed: signed,
			dim: dim};

	}

	// Set ident
	var elem;
	elem=decl.*[decl.*.length()-1];
	if (elem.name()=="id") type.ident=String(elem);
	else type.ident=String(elem..id[0]);

	if ((type.type=="struct" || type.type=="union") &&
		!type.struct_id)
		type.struct_id=type.ident;

	// Function declaration
	if (decl..fd!=undefined) {
		var params=[];
		var param;
		for each (param in decl..fd.pm.d) {
			var ptype=this.type(param);
			ptype.isConst=param["const"]!=undefined;
			params.push(ptype);
		}

		var callconv;
		if (decl..stdcall==undefined) callconv="cdecl";
		else callconv="stdcall";

		// undefined argument list
		if (params.length==0) {
			params=[]; //undefined; // no support for undef param list throughout yet
		}

		// void argument list
		if (params!=undefined && params.length==1 && params[0].type=="void" && params[0].dim.length==0) {
			params=[];
		}

		var ident=type.ident;
		delete type.ident;

		// Default int return
		if (type.type==undefined) {
			type={type:"int", size:0, signed: undefined, dim:[]};
		}

		// void return
		if (type.type=="void" && type.dim.length==0) {
			type=undefined;
		}
		var dim=[];
		this.getdim(decl..fd.*[0], dim);

		type={type: "function",
			callconv: callconv,
			rettype: type,
			params: params,
			dim: dim,
			ident: ident,
			elipsis: decl..fd.pm.elipsis!=undefined};
	}

	return type;
		
}

C.parsed_code.prototype.getdim=function(declor, dim) {
	if (declor.name()=="ptr") {
		var asterisk=declor.a;
		while (asterisk!=undefined) {
			dim.push(0);
			asterisk=asterisk.a;
		}
		if (declor.*[1]!=undefined) this.getdim(declor.*[1], dim);
	} else if (declor.name()=="ix") {
		if (declor.*[1]==undefined) {
			dim.push(0)
		} else {
			dim.push(eval(this.inner_eval(declor.*[1])));
		}
		this.getdim(declor.*[0], dim);
	} else if (declor.name()=="p") {
		this.getdim(declor.*[declor.*.length()-1], dim);
	}
}

C.parsed_code.prototype.inner_eval=function (expr) {
	if (expr.name()=="c") {
		return expr.toString();
	}
	if (expr.name()=="p") {
		return "("+this.inner_eval(expr.*[0])+")";
	}
	if (expr.name()=="op" && expr.@op=="sizeof" && expr.@type=="t") {
		return this.sizeof(expr.*[0]);
	}
	if (expr.name()=="op" && expr.@op=="sizeof" && expr.@type=="e") {
		return this.sizeof(this.vars[expr..id]);
	}
	if (expr.name()=="op" && expr.*.length()==1) {
		return expr.@op+this.inner_eval(expr.*[0]);
	}
	if (expr.name()=="op" && expr.*.length()==2) {
		return this.inner_eval(expr.*[0])+expr.@op+this.inner_eval(expr.*[1]);
	}
	if (expr.name()=="op" && expr.*.length()==3) {
		return this.inner_eval(expr.*[0])+"?"+this.inner_eval(expr.*[1])+":"+this.inner_eval(expr.*[2]);
	}
}

C.parsed_code.intsizes=[1,2,4,4,8,8]; // char, short, int, long, long long, __int64
C.parsed_code.floatsizes=[4,8]; // float, double
C.parsed_code.ptrsize=4; // any pointer

C.parsed_code.prototype.structsize=function(name) {
	var struct=this.structs[name];
	var decl;
	var sizeof=0;
	for each (decl in struct.struct.d) {
		sizeof+=this.sizeof(decl);
	}
	return sizeof;
}

C.parsed_code.prototype.unionsize=function(name) {
	var union=this.structs[name]; // unions are also stored in "structs".
	var decl;
	var sizeof=0;
	for each (decl in union.union.d) {
		var thissizeof=this.sizeof(decl);
		if (thissizeof>sizeof) sizeof=thissizeof;
	}
	return sizeof;
}

C.parsed_code.prototype.sizeof=function(decl) {
	var type=this.type(decl);
	var sizeof=0;
	var i;
	for (i in type.dim) {
		if (type.dim[i]==0) {
			sizeof=C.parsed_code.ptrsize;
			break;
		}
	}

	if (sizeof==0) {
		if (type.type=="int") sizeof=C.parsed_code.intsizes[type.size];
		if (type.type=="float") sizeof=C.parsed_code.floatsizes[type.size];
		if (type.type=="struct") sizeof=this.structsize(type.struct_id);
		if (type.type=="union") sizeof=this.unionsize(type.struct_id);
	}

	for (i in type.dim) {
		if (type.dim[i]==0) {
			break;
		} else {
			sizeof*=type.dim[i];
		}
	}

	return sizeof;
}

/*

	Tries to coerce a C macro into a JavaScript expression

*/

C.parsed_code.prototype.macro=function(macro) {
// Change 123L into 123
	var v=" "+macro.v+" ";
	v=v.replace(/([^a-zA-Z_0-9])([0-9]+)[lLuUfF]/g,"$1$2");
	v=v.replace(/([^a-zA-Z_0-9])(0x[0-9a-fA-F]+)[lLuU]/g,"$1$2");

// remove typecasts
    if (this.typecastregexp==undefined) {
		var types=["(unsigned +|signed +)?(long +|short +)?(int|char|long|short|signed|unsigned)", "double", "float"];
		for (var i in this.types) {
			types.push(i);
		}
		this.typecastregexp=new RegExp("\\( *(("+types.join(")|(")+")) *\\)","g");
	}

	v=v.replace(this.typecastregexp,"");

// alter references to global variables

    if (this.globvarregexp==undefined) {
		this.globvarregexp=[];
		var globvars=[];
		for each (var ptr in this.ptrs) {
			var type=this.type(ptr);
			if (type.type!="function") globvars.push(type.ident);
		}
		for (var i in globvars) {
			this.globvarregexp[i]=
				new RegExp("([^a-zA-Z_])("+globvars[i]+")([^a-zA-Z_0-9])","g");
		}
	}

	for each(var rx in this.globvarregexp)
		v=v.replace(rx,"$1$2.get()$3");

// change struct indexing

	v=v.replace(/->[ \t]*([a-zA-Z_][a-zA-Z_0-9]*)/g,"['$1']");

// remove L before string like in L"string"

	v=v.replace(/^([^"])*L"/,'$1"');

 // trim, don't use trim module because of build deps
	v=v.replace(/^ */,"").replace(/ *$/,"");


	if (macro.pm.length()) {
		var params=[];
		for each (var param in macro.pm)
			params.push(param);
		try {
			return new Function(params.join(","),
			"with (this, arguments) { return "+v+" }");
		} catch(x) {
			return;
		}
	} else if (v=="") {
		return "null";
	} else {
		return v;
	}

}