// Contains posix-specific functions

os.sep='/';
os.pathsep=':';
os.dlext='.so';

/*
listdir([path="."]) -> list_of_strings
Return a list containing the names of the entries in the directory.
 
        path: path of directory to list
 
The list is in arbitrary order.  It does not include the special
entries '.' and '..' even if they are present in the directory.
*/

os.listdir=function(path) {
	if (path==undefined) path=".";
	var d=_.opendir(path);
	if (!d) os.error("os.listdir");

	var ret=[];

	for(;;) {
		var e=_.readdir(d);
		if (e==null) break;
		var str=String.fromCharCode.apply(String,e.d_name);
		var len=str.indexOf(String.fromCharCode(0));
		str=str.substr(0,len);
		if (str!="." && str!="..") ret.push(str);
	}

	_.closedir(d);
	return ret;
}


/*
lstat(path) -> (mode,ino,dev,nlink,uid,gid,size,atime,mtime,ctime)

Like stat(path), but do not follow symbolic links.
*/

os.lstat=function(path) {
	var ret=_.New(_.struct.stat);
	if (_.call_lstat(path, ret)==-1) {
		return null;
	}
	return os.unistat(ret);
}

/*
fstat(fd) -> (mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime)

Like stat(), but for an open file descriptor.
*/

os.fstat=function(fd) {
	var ret=_.New(_.struct.stat);
	if (_.call_fstat(fd, ret)==-1) {
		return null;
	}
	return os.unistat(ret);
}

/*
stat(path) -> (mode,ino,dev=[major,minor],nlink,uid,gid,size,atime,mtime,ctime)

Perform a stat system call on the given path. Times are returned as date objects.
*/

os.stat=function(path) {
	var ret=_.New(_.struct.stat);
	if (_.call_stat(path, ret)==-1) {
		return null;
	}
	return os.unistat(ret);
}

/*
====isdir(path)====

Returns true if path exists and is a directory.
Otherwise, returns false
*/

os.isdir=function(path) {
	var ret=_.New(_.struct.stat);
	if (_.call_stat(path,ret)==-1) return false;
	if ((ret.st_mode & _.__S_IFMT) == _.__S_IFDIR) return true;
	return false;
}

/*
fstatvfs(fd) -> 
 (bsize, frsize, blocks, bfree, bavail, files, ffree, favail, flag, namemax)
Perform an fstatvfs system call on the given fd.
*/

os.fstatvfs=function(fd) {
	var buf=_.New(_.struct.statvfs);
	if (_.statvfs(fd,buf)) {
		os.error("os.fstatvfs");
	}
	var ret=buf.get();
	return ret;
}

/*
statvfs(path) -> 
 (bsize, frsize, blocks, bfree, bavail, files, ffree, favail, flag, namemax)
Perform a statvfs system call on the given path.
*/

os.statvfs=function(path) {
	var buf=_.New(_.struct.statvfs);
	if (_.statvfs(path,buf)) {
		os.error("os.statvfs");
	}
	var ret=buf.get();
	return ret;
}

/**/

os.unistat=function(buf) {
	var i;
	var j;
	var s=buf.get();
	var r={};
	for (i in s) {
		j=i.substr(3);
		switch (i) {
		case "st_atim":
		case "st_ctim":
		case "st_mtim":
			r[j+"e"]=new Date(s[i].tv_sec*1000);
			break;
		case "st_dev":
		case "st_rdev":
			r[j]=s[i].__val;
			break;
		default:
			r[j]=s[i];
		}
	}
	return r;
}

/*
mkstemp() -> file object

Create a temporary file with a directory entry.
*/

os.mkstemp=function(prefix) {
	if (prefix==undefined) prefix="/tmp/mkstemp";
	var arg=[];
	var i;
	prefix+="XXXXXX";
	for (i=0; i<prefix.length; i++)
		arg.push(prefix.charCodeAt(i));
	arg.push(0);
	var fd=_.mkstemp(arg);
	if (fd==-1) os.error("os.mkstemp");
	arg.pop();
	var ret=os.fdopen(fd, "w+b");
	ret.name=String.fromCharCode.apply(String,arg);
	return ret;
}

/*
tmpdir([prefix])

Return the name of a newly created, empty directory
*/

os.tmpdir=function(prefix) {
	if (prefix==undefined) prefix="/tmp/mkdtemp";
	var arg=[];
	var i;
	prefix+="XXXXXX";
	for (i=0; i<prefix.length; i++)
		arg.push(prefix.charCodeAt(i));
	arg.push(0);
	var fd=_.mkdtemp(arg);
	if (fd==null) os.error("os.mkdtemp");
	arg.pop();
	return String.fromCharCode.apply(String,arg);
}

os.isrelative=function(path) {
	return path[0]!=os.sep;
}

/*
errno() -> Number
Returns last error number
*/

os.errno=function() {
	return _.__errno_location().get();
}
